// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>


void sim_tof(struct device *in)
{
//<clean>
in->dt=in->tofdtstart;

carrier_count_reset(in);

in->go_time=FALSE;


struct istruct tofout;
inter_init(&tofout);

printf("Running tof. %e\n",in->externalv);
double V=in->Vstart;
int ittr=0;
//double total_steps=0;
int step=0;

solve_light(in,&(in->mylight),in->Psun,0.0);

ramp(in,in->Vapplied,in->externalv,0.1);
sim_externalv(in,in->externalv);
time_init(in);
device_timestep(in);

in->go_time=TRUE;
V=in->Vapplied;

in->time=-in->pulsestart*in->dt;
//int dump_count=0;
int tof_running=FALSE;
double np_out_tot=0.0;

do
{
if (step==in->pulsestart) carrier_count_reset(in);

	if (step<in->pulsestart)
	{
		solve_light(in,&(in->mylight),in->Psun,0.0);
		reset_npinit(in);
	}else
	if ((step>=in->pulsestart)&&(in->time<(in->toflaseroff)))
	{
		solve_light(in,&(in->mylight),in->Psun,in->toflaserpower);

	}else
	{
		if (tof_running==FALSE)
		{
			np_out_tot=0.0;
			tof_running=TRUE;
		}
		solve_light(in,&(in->mylight),in->Psun,0.0);
	}

	in->Vapplied=V;

	solve_all(in);

	#ifndef mindump
	printf("tof time=%e step=%d %.1e ",in->time,step,in->last_error);
	char out[100];
	sprintf(out,"V=%lf current = %e mA (%e A/m^2)\n",V,get_I(in)/1e-3,get_J(in));

	printdebug(out);
	#endif

	ittr++;

	plot_now(in,"plot");

	if (in->time>0.0) in->dt*=in->tofdtmull;

	inter_append(&tofout,in->dt,fabs(get_J_left(in)));

	device_timestep(in);
	step++;

	if (in->time>in->tofstop) break;

	if (in->stop==TRUE)
	{
		break;

	}


}while(1);


in->go_time=FALSE;

char outpath[200];

sprintf(outpath,"%s%s",in->outputpath,"tof.dat");
inter_sub(&tofout,tofout.data[tofout.len-1]-1e-6);
inter_save(&tofout,outpath);
inter_free(&tofout);

//</clean>
}
