// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>

void sim_tpv(struct device *in)
{
//<clean>
int i;
int tpv_running=FALSE;
in->dt=in->tpvdtstart;


carrier_count_reset(in);

in->go_time=FALSE;

struct istruct tpvvoc;
in->Vapplied=in->tpvvoc;
inter_init(&tpvvoc);

struct istruct gexout;
inter_init(&gexout);

struct istruct tpvnp;
inter_init(&tpvnp);

struct istruct only_n;
inter_init(&only_n);

struct istruct only_p;
inter_init(&only_p);



int step=0;




if (in->tpv_mode==tpv_mode_light)
{
printf("Running tpv. %e V\n",in->tpvvoc);
in->Psun=1000.0;
solve_light(in,&(in->mylight),in->Psun,0.0);
set_light_for_voc(in,in->tpvvoc);
in->Vapplied=in->tpvvoc;
}else
{
printf("Running tpv. %e W\n",in->Psun);
solve_light(in,&(in->mylight),in->Psun,0.0);
newton_sim_tpv(in);
}
printf("here\n");
if (in->simwire==TRUE)
{
	wire_init(in->Vapplied,1e-9);
}


//sim_externalv(in,in->in->tpvvoc);
time_init(in);
device_timestep(in);

in->go_time=TRUE;

in->time=-in->tpvpulsestart*in->dt;

FILE *out=fopen("./tpvpsun.dat","w");
fprintf(out,"%le\n",in->Psun);
fclose(out);

FILE *tpvn=fopen("./tpvn.dat","w");
fclose(tpvn);

tpvn=fopen("./tpv_cden.dat","w");
fprintf(tpvn,"%le\n",get_extracted_np(in));
fclose(tpvn);

FILE *srhout=fopen("./tpvsrh.dat","w");

do
{
	if (step<in->tpvpulsestart)
	{
		solve_light(in,&(in->mylight),in->Psun,0.0);
		reset_npinit(in);
	}else
	if ((step>=in->tpvpulsestart)&&(in->time<(in->tpvlaseroff)))
	{
		//
		solve_light(in,&(in->mylight),in->Psun,in->tpvlaserpower);
	}else
	{
		if (tpv_running==FALSE)
		{
			//
			tpv_running=TRUE;
		}
		solve_light(in,&(in->mylight),in->Psun,0.0);
	}

	//solve_all(in);
	if (in->simwire==FALSE)
	{
		newton_sim_tpv_fast(in);

	}else
	{
		wire_set_dt(in->dt);
		newton_sim_tpv_wire(in);
		wire_addtime();
	}

	#ifndef mindump
	printf("tpv time=%e step=%d %.1e ",in->time,step,in->last_error);
	char out[100];
	sprintf(out,"V=%lf current = %e mA (%e A/m^2)\n",in->Vapplied,get_I(in)/1e-3,get_J(in));

	printdebug(out);
	#endif


	if (tpv_running==TRUE)
	{

		double a=0.0;
		double b=0.0;
		double c=0.0;
		double d=0.0;

		for (i=0;i<in->ymeshpoints;i++)
		{
			a+=in->nrelax[i];
			b+=in->ntrap_to_p[i];
			c+=in->prelax[i];
			d+=in->ptrap_to_n[i];
		}
		a/=in->ymeshpoints;
		b/=in->ymeshpoints;
		c/=in->ymeshpoints;
		d/=in->ymeshpoints;

		fprintf(srhout,"%le %le %le %le %le\n",in->time,a,b,c,d);


		if (in->simwire==FALSE)
		{
			inter_append(&tpvvoc,in->time,get_V_TPV(in));
		}else
		{
			inter_append(&tpvvoc,in->time,wire_get_voltage_at_end());
		}
		inter_append(&gexout,in->time,in->Gn[0]);

		FILE *tpvn=fopen("./tpvn.dat","a");
		fprintf(tpvn,"%le %le %le %le %le\n",in->time,get_free_n_charge(in),get_free_p_charge(in),get_n_trapped_charge(in),get_p_trapped_charge(in));
		fclose(tpvn);

		inter_append(&tpvnp,in->time,get_extracted_np(in));
		inter_append(&only_n,in->time,get_extracted_n(in)/2.0);
		inter_append(&only_p,in->time,get_extracted_p(in)/2.0);
	}

	plot_now(in,"plot");

	if (in->time>in->tpvlaseroff) in->dt*=in->tpvdtmull;

	/*if (step==in->tpvpulsestart)
	{
		in->dt=in->tpvdtstart;
	}*/
	step++;
	device_timestep(in);



	if (in->time>in->tpvstop) break;

	if (in->stop==TRUE)
	{
		break;

	}

}while(1);

fclose(srhout);

in->go_time=FALSE;

char outpath[200];



inter_sub(&tpvnp,tpvnp.data[tpvnp.len-1]);
sprintf(outpath,"%s%s",in->outputpath,"tpv_np.dat");
inter_save(&tpvnp,outpath);
inter_free(&tpvnp);

inter_sub(&tpvvoc,tpvvoc.data[tpvvoc.len-1]);

inter_mod(&tpvvoc);

sprintf(outpath,"%s%s",in->outputpath,"tpv_voc.dat");
inter_save(&tpvvoc,outpath);
inter_free(&tpvvoc);

inter_sub(&only_n,only_n.data[only_n.len-1]);
sprintf(outpath,"%s%s",in->outputpath,"tpv_n.dat");
inter_save(&only_n,outpath);
inter_free(&only_n);

inter_sub(&only_p,only_p.data[only_p.len-1]);
sprintf(outpath,"%s%s",in->outputpath,"tpv_p.dat");
inter_save(&only_p,outpath);
inter_free(&only_p);

sprintf(outpath,"%s%s",in->outputpath,"gex.dat");
inter_save(&gexout,outpath);
inter_free(&gexout);
//</clean>
}

